"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

const connect = async (imports, pool, options) => {
  let {
    tablePrefix,
    databaseName,
    buildMode,
    useSqs,
    ...connectionOptions
  } = options;

  if (databaseName == null || databaseName.constructor !== String) {
    throw new Error(`Wrong database name: ${databaseName}`);
  }

  if (tablePrefix != null && tablePrefix.constructor !== String) {
    throw new Error(`Wrong table prefix: ${tablePrefix}`);
  } else if (tablePrefix == null) {
    tablePrefix = '';
  }

  if (buildMode != null && !['plv8-internal', 'plv8-external', 'plv8', 'nodejs', 'auto'].includes(buildMode)) {
    throw new Error(`Wrong build mode: ${buildMode}`);
  } else if (buildMode == null) {
    buildMode = 'auto';
  }

  if (useSqs != null && useSqs.constructor !== Boolean) {
    throw new Error(`Wrong sqs usage flag: ${useSqs}`);
  } else if (useSqs == null) {
    useSqs = false;
  }

  const connectionErrorsMap = new WeakMap();
  const connectPromiseMap = new WeakMap();

  const establishConnection = async () => {
    if (pool.connection != null) {
      const connection = pool.connection;
      await connectPromiseMap.get(connection);
      return connection;
    }

    const connection = new imports.Postgres(connectionOptions);
    pool.connection = connection;
    connectPromiseMap.set(connection, (async () => {
      await Promise.resolve();
      connectionErrorsMap.set(connection, []);

      try {
        connection.on('error', error => {
          var _connectionErrorsMap$;

          (_connectionErrorsMap$ = connectionErrorsMap.get(connection)) === null || _connectionErrorsMap$ === void 0 ? void 0 : _connectionErrorsMap$.push(error);
        });
        await connection.connect();
        await connection.query('SELECT 0 AS "defunct"');
      } catch (error) {
        var _connectionErrorsMap$2;

        (_connectionErrorsMap$2 = connectionErrorsMap.get(connection)) === null || _connectionErrorsMap$2 === void 0 ? void 0 : _connectionErrorsMap$2.push(error);
      }
    })());
    await connectPromiseMap.get(connection);
    return connection;
  };

  const maybeThrowConnectionErrors = async connection => {
    var _connectionErrorsMap$3;

    const connectionErrors = (_connectionErrorsMap$3 = connectionErrorsMap.get(connection)) !== null && _connectionErrorsMap$3 !== void 0 ? _connectionErrorsMap$3 : [];

    if (connectionErrors.length > 0) {
      let summaryError = connectionErrors[0];

      if (connectionErrors.length > 1) {
        summaryError = new Error(connectionErrors.map(({
          message
        }) => message).join('\n'));
        summaryError.stack = connectionErrors.map(({
          stack
        }) => stack).join('\n');
      }

      if (pool.connection === connection) {
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        pool.connection = null;
      }

      try {
        await connection.end();
      } catch (err) {}

      throw summaryError;
    }
  };

  const initialConnection = await establishConnection();
  await maybeThrowConnectionErrors(initialConnection);

  const inlineLedgerRunQuery = async (sql, passthroughRuntimeErrors = false) => {
    let result = null;

    for (;;) {
      try {
        const connection = await establishConnection();
        await maybeThrowConnectionErrors(connection);
        result = await connection.query(sql);
        break;
      } catch (error) {
        if (pool.activePassthrough) {
          if (imports.PassthroughError.isPassthroughError(error, !!passthroughRuntimeErrors)) {
            imports.PassthroughError.maybeThrowPassthroughError(error, !!passthroughRuntimeErrors);
          } else {
            throw error;
          }
        } else {
          if (imports.PassthroughError.isPassthroughError(error, false)) {
            await new Promise(resolve => setTimeout(resolve, 100));
          } else {
            throw error;
          }
        }
      }
    }

    let rows = null;

    if (result != null && Array.isArray(result.rows)) {
      rows = JSON.parse(JSON.stringify(result.rows));
    }

    return rows;
  };

  Object.assign(pool, {
    schemaName: databaseName,
    tablePrefix,
    inlineLedgerRunQuery,
    connection: initialConnection,
    activePassthrough: false,
    buildMode,
    useSqs,
    ...imports
  });
};

var _default = connect;
exports.default = _default;
//# sourceMappingURL=connect.js.map